#!/usr/bin/env python
#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

""" Tool to automatically format source code in Nuitka style.

"""

from __future__ import print_function

import os
import re
import shutil
import sys

from redbaron import RedBaron  # @UnresolvedImport


def autoformat(filename):
    # All the complexity in one place, pylint: disable=R0912,R0915
    print("Consider", filename, end = ": ")

    old_code = open(filename, 'r').read()

    red = RedBaron(old_code.rstrip()+'\n')

    def updateCall(call_node):
        max_len = 0
        for argument in call_node:
            if argument.type == "argument_generator_comprehension":
                return

            if hasattr(argument, "target") and argument.target is not None:
                key = argument.target.value
            else:
                key = None

            if key is not None:
                max_len = max(max_len, len(key))

        if '\n' not in call_node.second_formatting.dumps():
            del call_node.second_formatting[:]
            del call_node.third_formatting[:]

        for argument in call_node:
            if hasattr(argument, "target") and argument.target is not None:
                key = argument.target.value
            else:
                key = None

            if key is not None:
                if not argument.second_formatting:
                    argument.second_formatting = ' '

                if '\n' in str(call_node.second_formatting):
                    if len(argument.first_formatting) > 0:
                        spacing = argument.first_formatting[0].value
                    else:
                        spacing = ""

                    if len(key)+len(spacing) != max_len + 1:
                        argument.first_formatting = ' ' * (max_len - len(key) + 1)
                else:
                    argument.first_formatting = ' '
            else:
                if '\n' not in str(call_node.second_formatting):
                    if argument.value.type in ("string", "binary_string", "raw_string"):
                        argument.value.second_formatting = ""

    def updateTuple(tuple_node):
        if '\n' not in str(tuple_node.dumps()):
            tuple_node.second_formatting = ""
            tuple_node.third_formatting = ""

            if tuple_node.type == "tuple" and tuple_node.with_parenthesis:
                if len(tuple_node.value.node_list) > 0:
                    if tuple_node.value.node_list[-1].type not in ("yield_atom",):
                        tuple_node.value.node_list[-1].second_formatting = ""

            for argument in tuple_node.value:
                if argument.type in ("string", "binary_string", "raw_string"):
                    argument.second_formatting = ""

    def updateString(string_node):
        # Skip doc strings for now.
        if not hasattr(node.parent, "type") or \
           string_node.parent.type in ("class", "def", None):
            return

        value = string_node.value

        def isQuotedWith(quote):
            return value.startswith(quote) and value.endswith(quote)

        quote = None # For PyLint.
        for quote in "'''", '"""', "'", '"':
            if isQuotedWith(quote):
                break
        else:
            sys.exit("Error, quote not understood.")

        real_value = value[len(quote):-len(quote)]
        assert quote + real_value + quote == value

        if '\n' not in real_value:
            # Single characters, should be quoted with "'"
            if len(eval(value)) == 1: # pylint: disable=W0123
                if real_value != "'":
                    string_node.value = "'" + real_value + "'"
            else:
                if '"' not in real_value:
                    string_node.value = '"' + real_value + '"'

    def updateDefNode(def_node):
        # This is between "def" and function name.
        def_node.first_formatting = ' '

        # This is after the opening/closing brace, we don't want it there.
        def_node.third_formatting = ""
        def_node.fourth_formatting = ""

        # This is to insert/remove spaces or new lines, depending on line length
        # so far, but is not functional at all.
        for argument_node in def_node.arguments:
            argument_node.first_formatting = ' '
            argument_node.second_formatting = ' '

    def updateCommentNode(comment_node):

        if "pylint:" in str(comment_node.value):
            def replacer(part):
                return part.group(1) + ','.join(sorted(part.group(2).split(',')))

            new_value = re.sub(r"(pylint\: disable=)(.*)", replacer, str(comment_node.value), flags = re.M)
            comment_node.value = new_value

    for node in red.find_all("CallNode"):
        try:
            updateCall(node)
        except Exception:
            print("Problem with", node)
            node.help(deep = True, with_formatting = True)
            raise

    for node in red.find_all("TupleNode"):
        try:
            updateTuple(node)
        except Exception:
            print("Problem with", node)
            node.help(deep = True, with_formatting = True)
            raise

    for node in red.find_all("ListNode"):
        try:
            updateTuple(node)
        except Exception:
            print("Problem with", node)
            node.help(deep = True, with_formatting = True)
            raise

    for node in red.find_all("SetNode"):
        try:
            updateTuple(node)
        except Exception:
            print("Problem with", node)
            node.help(deep = True, with_formatting = True)
            raise

    for node in red.find_all("StringNode"):
        try:
            updateString(node)
        except Exception:
            print("Problem with", node)
            node.help(deep = True, with_formatting = True)
            raise

    for node in red.find_all("DefNode"):
        try:
            updateDefNode(node)
        except Exception:
            print("Problem with", node)
            node.help(deep = True, with_formatting = True)
            raise

    for node in red.find_all("CommentNode"):
        try:
            updateCommentNode(node)
        except Exception:
            print("Problem with", node)
            node.help(deep = True, with_formatting = True)
            raise

    new_code = red.dumps()

    if new_code != old_code:

        new_name = filename + ".new"

        with open(new_name, 'w') as source_code:
            source_code.write(red.dumps())

        # There is no way to safely replace a file on Windows, but lets try on Linux
        # at least.
        old_stat = os.stat(filename)

        try:
            os.rename(new_name, filename)
        except OSError:
            shutil.copy(new_name, filename)

        os.chmod(filename, old_stat.st_mode)

        print("updated.")
    else:
        print("OK.")



#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Many tools work on Nuitka sources and need to find the files.

"""

import os


def addFromDirectory(path, blacklist):
    for dirpath, dirnames, filenames in os.walk(path):
        dirnames.sort()

        if "inline_copy" in dirnames:
            dirnames.remove("inline_copy")

        filenames.sort()

        for filename in filenames:
            fullpath = os.path.join(dirpath, filename)

            # Ignore links
            if os.path.islink(fullpath):
                continue

            # Skip temporary files from flymake mode of Emacs.
            if filename.endswith("_flymake.py"):
                continue
            # Skip temporary files from unsaved files of Emacs.
            if filename.startswith(".#"):
                continue

            # Python files only. TODO: Provided this from the outside.
            if not filename.endswith(".py"):
                line = open(fullpath).readline()
                if not line.startswith("#!") or "python" not in line:
                    continue

            if filename in blacklist:
                continue

            yield fullpath


def scanTargets(positional_args, blacklist = ()):
    for positional_arg in positional_args:
        if os.path.isdir(positional_arg):
            for value in addFromDirectory(positional_arg, blacklist):
                yield value
        else:
            yield positional_arg



#!/usr/bin/env python
#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

""" Main program for PyLint checker tool.

"""

from __future__ import print_function

import os
import sys
from optparse import OptionParser

# Unchanged, running from checkout, use the parent directory, the nuitka
# package ought be there.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            ".."
        )
    )
)

from nuitka.tools.Basics import goHome, addPYTHONPATH # isort:skip
from nuitka.tools.ScanSources import scanTargets # isort:skip
from nuitka.tools.pylint.PyLint import executePyLint # isort:skip

def main():
    goHome()

    # So PyLint finds nuitka package.
    addPYTHONPATH(os.getcwd())

    parser = OptionParser()

    parser.add_option(
        "--show-todos", "--todos",
        action  = "store_true",
        dest    = "todos",
        default = False,
        help    = """\
        Default is %default."""
    )

    parser.add_option(
        "--verbose",
        action  = "store_true",
        dest    = "verbose",
        default = False,
        help    = """\
        Default is %default."""
    )

    options, positional_args = parser.parse_args()

    if not positional_args:
        positional_args = ["bin", "nuitka"]

    print("Working on:", positional_args)

    blacklist = (
        "oset.py",
        "odict.py",
        "SyntaxHighlighting.py",
    )


    found = False
    for filename in scanTargets(positional_args, blacklist):
        executePyLint(filename, options.todos, options.verbose)
        found = True

    if not found:
        sys.exit("No files found.")

if __name__ == "__main__":
    main()



#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Code templates for frames of all kinds.

"""

template_frame_guard_cache_decl = """\
static PyFrameObject *cache_%(frame_identifier)s = NULL;
"""

template_frame_guard_frame_decl = """\
PyFrameObject *%(frame_identifier)s;
"""

# Frame in a function
template_frame_guard_full_block = """\
MAKE_OR_REUSE_FRAME( cache_%(frame_identifier)s, %(code_identifier)s, %(module_identifier)s );
%(frame_identifier)s = cache_%(frame_identifier)s;

// Push the new frame as the currently active one.
pushFrameStack( %(frame_identifier)s );

// Mark the frame object as in use, ref count 1 will be up for reuse.
Py_INCREF( %(frame_identifier)s );
assert( Py_REFCNT( %(frame_identifier)s ) == 2 ); // Frame stack

#if PYTHON_VERSION >= 340
%(frame_identifier)s->f_executing += 1;
#endif

// Framed code:
%(codes)s

#if %(needs_preserve)d
RESTORE_FRAME_EXCEPTION( %(frame_identifier)s );
#endif
// Put the previous frame back on top.
popFrameStack();
#if PYTHON_VERSION >= 340
%(frame_identifier)s->f_executing -= 1;
#endif
Py_DECREF( %(frame_identifier)s );
goto %(no_exception_exit)s;
"""

template_frame_guard_full_return_handler = """\
%(frame_return_exit)s:;
#if %(needs_preserve)d
RESTORE_FRAME_EXCEPTION( %(frame_identifier)s );
#endif
popFrameStack();
#if PYTHON_VERSION >= 340
%(frame_identifier)s->f_executing -= 1;
#endif
Py_DECREF( %(frame_identifier)s );
goto %(return_exit)s;
"""

template_frame_guard_full_exception_handler = """\
%(frame_exception_exit)s:;
#if %(needs_preserve)d
RESTORE_FRAME_EXCEPTION( %(frame_identifier)s );
#endif

{
    bool needs_detach = false;

    if ( exception_tb == NULL )
    {
        exception_tb = %(tb_making)s;
        needs_detach = true;
    }
    else if ( exception_lineno != -1 )
    {
        PyTracebackObject *traceback_new = MAKE_TRACEBACK( %(frame_identifier)s, exception_lineno );
        traceback_new->tb_next = exception_tb;
        exception_tb = traceback_new;

        needs_detach = true;
    }

    if (needs_detach)
    {
%(store_frame_locals)s

        detachFrame( exception_tb, %(frame_locals_name)s );
    }
}

popFrameStack();

#if PYTHON_VERSION >= 340
%(frame_identifier)s->f_executing -= 1;
#endif
Py_DECREF( %(frame_identifier)s );

// Return the error.
goto %(parent_exception_exit)s;
"""

# Frame for a module. TODO: Use it for functions called only once.
# TODO: The once guard need not take a reference count in its frame class.
template_frame_guard_once = """\
// Frame without reuse.
%(frame_identifier)s = MAKE_MODULE_FRAME( %(code_identifier)s, %(module_identifier)s );

// Push the new frame as the currently active one, and we should be exclusively
// owning it.
pushFrameStack( %(frame_identifier)s );
assert( Py_REFCNT( %(frame_identifier)s ) == 1 );

#if PYTHON_VERSION >= 340
%(frame_identifier)s->f_executing += 1;
#endif

// Framed code:
%(codes)s

// Restore frame exception if necessary.
#if %(needs_preserve)d
RESTORE_FRAME_EXCEPTION( %(frame_identifier)s );
#endif
popFrameStack();

assertFrameObject( %(frame_identifier)s );
Py_DECREF( %(frame_identifier)s );

goto %(no_exception_exit)s;
%(frame_exception_exit)s:;
#if %(needs_preserve)d
RESTORE_FRAME_EXCEPTION( %(frame_identifier)s );
#endif

if ( exception_tb == NULL )
{
    exception_tb = %(tb_making)s;
}
else if ( exception_tb->tb_frame != %(frame_identifier)s )
{
    PyTracebackObject *traceback_new = MAKE_TRACEBACK( %(frame_identifier)s, exception_lineno );
    traceback_new->tb_next = exception_tb;
    exception_tb = traceback_new;
}

// Put the previous frame back on top.
popFrameStack();

#if PYTHON_VERSION >= 340
%(frame_identifier)s->f_executing -= 1;
#endif
Py_DECREF( %(frame_identifier)s );

// Return the error.
goto %(parent_exception_exit)s;
%(no_exception_exit)s:;"""

# Frame in a generator
template_frame_guard_generator = """\
MAKE_OR_REUSE_FRAME( %(frame_cache_identifier)s, %(code_identifier)s, %(module_identifier)s );
generator->m_frame = %(frame_cache_identifier)s;
Py_INCREF( generator->m_frame );

#if PYTHON_VERSION >= 340
generator->m_frame->f_gen = (PyObject *)generator;
#endif

Py_CLEAR( generator->m_frame->f_back );

generator->m_frame->f_back = PyThreadState_GET()->frame;
Py_INCREF( generator->m_frame->f_back );

PyThreadState_GET()->frame = generator->m_frame;
Py_INCREF( generator->m_frame );

#if PYTHON_VERSION >= 340
generator->m_frame->f_executing += 1;
#endif

#if PYTHON_VERSION >= 300
// Accept currently existing exception as the one to publish again when we
// yield or yield from.

PyThreadState *thread_state = PyThreadState_GET();

generator->m_frame->f_exc_type = thread_state->exc_type;
if ( generator->m_frame->f_exc_type == Py_None ) generator->m_frame->f_exc_type = NULL;
Py_XINCREF( generator->m_frame->f_exc_type );
generator->m_frame->f_exc_value = thread_state->exc_value;
Py_XINCREF( generator->m_frame->f_exc_value );
generator->m_frame->f_exc_traceback = thread_state->exc_traceback;
Py_XINCREF( generator->m_frame->f_exc_traceback );
#endif

// Framed code:
%(codes)s

#if PYTHON_VERSION >= 340
generator->m_frame->f_executing -= 1;
#endif

#if PYTHON_VERSION >= 300
Py_CLEAR( generator->m_frame->f_exc_type );
Py_CLEAR( generator->m_frame->f_exc_value );
Py_CLEAR( generator->m_frame->f_exc_traceback );
#endif

Py_DECREF( generator->m_frame );
goto %(no_exception_exit)s;
"""

# Frame in a coroutine
template_frame_guard_coroutine = """\
MAKE_OR_REUSE_FRAME( %(frame_cache_identifier)s, %(code_identifier)s, %(module_identifier)s );
coroutine->m_frame = %(frame_cache_identifier)s;
Py_INCREF( coroutine->m_frame );

coroutine->m_frame->f_gen = (PyObject *)coroutine;

Py_CLEAR( coroutine->m_frame->f_back );

coroutine->m_frame->f_back = PyThreadState_GET()->frame;
Py_INCREF( coroutine->m_frame->f_back );

PyThreadState_GET()->frame = coroutine->m_frame;
Py_INCREF( coroutine->m_frame );

coroutine->m_frame->f_executing += 1;

// Accept currently existing exception as the one to publish again when we
// yield or yield from.

PyThreadState *thread_state = PyThreadState_GET();

coroutine->m_frame->f_exc_type = thread_state->exc_type;
if ( coroutine->m_frame->f_exc_type == Py_None ) coroutine->m_frame->f_exc_type = NULL;
Py_XINCREF( coroutine->m_frame->f_exc_type );
coroutine->m_frame->f_exc_value = thread_state->exc_value;
Py_XINCREF( coroutine->m_frame->f_exc_value );
coroutine->m_frame->f_exc_traceback = thread_state->exc_traceback;
Py_XINCREF( coroutine->m_frame->f_exc_traceback );

// Framed code:
%(codes)s

coroutine->m_frame->f_executing -= 1;

Py_CLEAR( coroutine->m_frame->f_exc_type );
Py_CLEAR( coroutine->m_frame->f_exc_value );
Py_CLEAR( coroutine->m_frame->f_exc_traceback );

Py_DECREF( coroutine->m_frame );
goto %(no_exception_exit)s;
"""


# TODO: This cannot happen, can it?
template_frame_guard_generator_return_handler = """\
%(frame_return_exit)s:;

#if PYTHON_VERSION >= 300
Py_CLEAR( %(frame_identifier)s->f_exc_type );
Py_CLEAR( %(frame_identifier)s->f_exc_value );
Py_CLEAR( %(frame_identifier)s->f_exc_traceback );
#endif

Py_DECREF( %(frame_identifier)s );
goto %(return_exit)s;
"""


template_frame_guard_generator_exception_handler = """\
%(frame_exception_exit)s:;

// If it's not an exit exception, consider and create a traceback for it.
if ( !EXCEPTION_MATCH_GENERATOR( exception_type ) )
{
    int needs_detach = false;

    if ( exception_tb == NULL )
    {
        exception_tb = %(tb_making)s;
        needs_detach = true;
    }
    else if ( exception_tb->tb_frame != %(frame_identifier)s )
    {
        PyTracebackObject *traceback_new = MAKE_TRACEBACK( %(frame_identifier)s, exception_lineno );
        traceback_new->tb_next = exception_tb;
        exception_tb = traceback_new;

        needs_detach = true;
    }

    if (needs_detach)
    {
%(store_frame_locals)s

        detachFrame( exception_tb, %(frame_locals_name)s );
    }
}

#if PYTHON_VERSION >= 300
Py_CLEAR( %(frame_identifier)s->f_exc_type );
Py_CLEAR( %(frame_identifier)s->f_exc_value );
Py_CLEAR( %(frame_identifier)s->f_exc_traceback );
#endif

Py_DECREF( %(frame_identifier)s );
// Return the error.
goto %(parent_exception_exit)s;
"""

from . import TemplateDebugWrapper # isort:skip
TemplateDebugWrapper.checkDebug(globals())



#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" The code generation.

No language specifics at all are supposed to be present here. Instead it is
using primitives from the given generator to build code sequences (list of
strings).

As such this is the place that knows how to take a condition and two code
branches and make a code block out of it. But it doesn't contain any target
language syntax.
"""

from nuitka.__past__ import iterItems
from nuitka.PythonVersions import python_version

from . import Contexts, Emission
from .AttributeCodes import (
    generateAssignmentAttributeCode,
    generateAttributeLookupCode,
    generateAttributeLookupSpecialCode,
    generateBuiltinGetattrCode,
    generateBuiltinHasattrCode,
    generateBuiltinSetattrCode,
    generateDelAttributeCode
)
from .BranchCodes import generateBranchCode
from .BuiltinCodes import (
    generateBuiltinAnonymousRefCode,
    generateBuiltinBinCode,
    generateBuiltinBoolCode,
    generateBuiltinBytearrayCode,
    generateBuiltinComplexCode,
    generateBuiltinFloatCode,
    generateBuiltinHexCode,
    generateBuiltinOctCode,
    generateBuiltinOpenCode,
    generateBuiltinRange1Code,
    generateBuiltinRange2Code,
    generateBuiltinRange3Code,
    generateBuiltinRefCode,
    generateBuiltinSum1Code,
    generateBuiltinSum2Code,
    generateBuiltinType1Code,
    generateBuiltinType3Code,
    generateBuiltinXrange1Code,
    generateBuiltinXrange2Code,
    generateBuiltinXrange3Code
)
from .CallCodes import generateCallCode, getCallsCode, getCallsDecls
from .ClassCodes import (
    generateBuiltinIsinstanceCode,
    generateBuiltinSuperCode,
    generateSelectMetaclassCode
)
from .ComparisonCodes import generateComparisonExpressionCode
from .ConditionalCodes import (
    generateConditionalAndOrCode,
    generateConditionalCode
)
from .ConstantCodes import (
    generateConstantEllipsisReferenceCode,
    generateConstantFalseReferenceCode,
    generateConstantNoneReferenceCode,
    generateConstantReferenceCode,
    generateConstantTrueReferenceCode
)
from .CoroutineCodes import (
    generateAsyncIterCode,
    generateAsyncNextCode,
    generateAsyncWaitCode,
    generateMakeCoroutineObjectCode,
    getCoroutineObjectCode
)
from .DictCodes import (
    generateBuiltinDictCode,
    generateDictionaryCreationCode,
    generateDictOperationGetCode,
    generateDictOperationInCode,
    generateDictOperationRemoveCode,
    generateDictOperationSetCode,
    generateDictOperationUpdateCode
)
from .EvalCodes import (
    generateBuiltinCompileCode,
    generateEvalCode,
    generateExecCode,
    generateExecfileCode,
    generateLocalsDictSyncCode
)
from .ExceptionCodes import (
    generateBuiltinMakeExceptionCode,
    generateExceptionCaughtTracebackCode,
    generateExceptionCaughtTypeCode,
    generateExceptionCaughtValueCode,
    generateExceptionPublishCode,
    generateExceptionRefCode
)
from .ExpressionCodes import (
    generateExpressionOnlyCode,
    generateSideEffectsCode
)
from .FrameCodes import (
    generateFramePreserveExceptionCode,
    generateFrameRestoreExceptionCode
)
from .FunctionCodes import (
    generateFunctionCallCode,
    generateFunctionCreationCode,
    generateFunctionDeclCode,
    generateFunctionOutlineCode,
    getExportScopeCode,
    getFunctionCode,
    getFunctionDirectDecl
)
from .GeneratorCodes import (
    generateMakeGeneratorObjectCode,
    getGeneratorObjectCode
)
from .GlobalsLocalsCodes import (
    generateBuiltinDir1Code,
    generateBuiltinGlobalsCode,
    generateBuiltinLocalsCode,
    generateBuiltinVarsCode,
    generateSetLocalsCode
)
from .Helpers import (
    generateStatementSequenceCode,
    setExpressionDispatchDict,
    setStatementDispatchDict
)
from .IdCodes import generateBuiltinHashCode, generateBuiltinIdCode
from .ImportCodes import (
    generateBuiltinImportCode,
    generateImportModuleCode,
    generateImportModuleHardCode,
    generateImportNameCode,
    generateImportStarCode
)
from .IntegerCodes import generateBuiltinIntCode, generateBuiltinLongCode
from .IteratorCodes import (
    generateBuiltinIter1Code,
    generateBuiltinIter2Code,
    generateBuiltinLenCode,
    generateBuiltinNext1Code,
    generateBuiltinNext2Code,
    generateSpecialUnpackCode,
    generateUnpackCheckCode
)
from .ListCodes import (
    generateBuiltinListCode,
    generateListCreationCode,
    generateListOperationAppendCode,
    generateListOperationExtendCode,
    generateListOperationPopCode
)
from .LoaderCodes import getMetapathLoaderBodyCode
from .LoopCodes import (
    generateLoopBreakCode,
    generateLoopCode,
    generateLoopContinueCode
)
from .ModuleCodes import (
    generateModuleFileAttributeCode,
    getModuleCode,
    getModuleValues
)
from .OperationCodes import (
    generateOperationBinaryCode,
    generateOperationUnaryCode
)
from .PrintCodes import generatePrintNewlineCode, generatePrintValueCode
from .RaisingCodes import generateRaiseCode
from .ReturnCodes import (
    generateGeneratorReturnCode,
    generateReturnCode,
    generateReturnedValueRefCode
)
from .SetCodes import (
    generateBuiltinSetCode,
    generateSetCreationCode,
    generateSetLiteralCreationCode,
    generateSetOperationAddCode,
    generateSetOperationUpdateCode
)
from .SliceCodes import (
    generateAssignmentSliceCode,
    generateBuiltinSliceCode,
    generateDelSliceCode,
    generateSliceLookupCode
)
from .StringCodes import (
    generateBuiltinAsciiCode,
    generateBuiltinChrCode,
    generateBuiltinFormatCode,
    generateBuiltinOrdCode,
    generateBuiltinStrCode,
    generateBuiltinUnicodeCode,
    generateStringContenationCode
)
from .SubscriptCodes import (
    generateAssignmentSubscriptCode,
    generateDelSubscriptCode,
    generateSubscriptLookupCode
)
from .TryCodes import generateTryCode
from .TupleCodes import generateBuiltinTupleCode, generateTupleCreationCode
from .VariableCodes import (
    generateAssignmentVariableCode,
    generateDelVariableCode,
    generateVariableReferenceCode,
    generateVariableReleaseCode
)
from .YieldCodes import generateYieldCode, generateYieldFromCode

_generated_functions = {}


def generateFunctionBodyCode(function_body, context):
    function_identifier = function_body.getCodeName()

    if function_identifier in _generated_functions:
        return _generated_functions[function_identifier]

    # TODO: Generate both codes, and base direct/etc. decisions on context.

    if function_body.isExpressionGeneratorObjectBody():
        function_context = Contexts.PythonGeneratorObjectContext(
            parent   = context,
            function = function_body
        )
    elif function_body.isExpressionCoroutineObjectBody():
        function_context = Contexts.PythonCoroutineObjectContext(
            parent   = context,
            function = function_body
        )
    elif function_body.isExpressionClassBody():
        function_context = Contexts.PythonFunctionDirectContext(
            parent   = context,
            function = function_body
        )
    elif function_body.needsCreation():
        function_context = Contexts.PythonFunctionCreatedContext(
            parent   = context,
            function = function_body
        )
    else:
        function_context = Contexts.PythonFunctionDirectContext(
            parent   = context,
            function = function_body
        )

    function_codes = Emission.SourceCodeCollector()

    generateStatementSequenceCode(
        statement_sequence = function_body.getBody(),
        allow_none         = True,
        emit               = function_codes,
        context            = function_context
    )

    needs_exception_exit = function_body.mayRaiseException(BaseException)

    if function_body.isExpressionGeneratorObjectBody():
        function_code = getGeneratorObjectCode(
            context                = function_context,
            function_identifier    = function_identifier,
            user_variables         = function_body.getUserLocalVariables(),
            temp_variables         = function_body.getTempVariables(),
            function_codes         = function_codes.codes,
            needs_exception_exit   = needs_exception_exit,
            needs_generator_return = function_body.needsGeneratorReturnExit()
        )
    elif function_body.isExpressionCoroutineObjectBody():
        function_code = getCoroutineObjectCode(
            context                = function_context,
            function_identifier    = function_identifier,
            user_variables         = function_body.getUserLocalVariables(),
            temp_variables         = function_body.getTempVariables(),
            function_codes         = function_codes.codes,
            needs_exception_exit   = needs_exception_exit,
            needs_generator_return = function_body.needsGeneratorReturnExit()
        )
    elif function_body.isExpressionClassBody():
        function_code = getFunctionCode(
            context              = function_context,
            function_identifier  = function_identifier,
            parameters           = None,
            closure_variables    = function_body.getClosureVariables(),
            user_variables       = function_body.getUserLocalVariables(),
            temp_variables       = function_body.getTempVariables(),
            function_codes       = function_codes.codes,
            function_doc         = function_body.getDoc(),
            needs_exception_exit = needs_exception_exit,
            file_scope           = getExportScopeCode(
                cross_module = False
            )
        )
    else:
        parameters = function_body.getParameters()

        function_code = getFunctionCode(
            context              = function_context,
            function_identifier  = function_identifier,
            parameters           = parameters,
            closure_variables    = function_body.getClosureVariables(),
            user_variables       = function_body.getUserLocalVariables(),
            temp_variables       = function_body.getTempVariables(),
            function_codes       = function_codes.codes,
            function_doc         = function_body.getDoc(),
            needs_exception_exit = needs_exception_exit,
            file_scope           = getExportScopeCode(
                cross_module = function_body.isCrossModuleUsed()
            )
        )

    return function_code, function_context



def prepareModuleCode(global_context, module, module_name):
    # As this not only creates all modules, but also functions, it deals
    # also with its functions.

    assert module.isCompiledPythonModule(), module

    context = Contexts.PythonModuleContext(
        module         = module,
        module_name    = module_name,
        code_name      = module.getCodeName(),
        filename       = module.getFilename(),
        global_context = global_context
    )

    context.setExceptionEscape("module_exception_exit")

    statement_sequence = module.getBody()

    codes = Emission.SourceCodeCollector()

    generateStatementSequenceCode(
        statement_sequence = statement_sequence,
        emit               = codes,
        allow_none         = True,
        context            = context,
    )

    function_decl_codes = []
    function_body_codes = []

    for function_body in module.getUsedFunctions():
        function_code, function_context = generateFunctionBodyCode(
            function_body = function_body,
            context       = context
        )

        assert type(function_code) is str, type(function_code)

        function_body_codes.append(function_code)

        function_decl = generateFunctionDeclCode(
            function_body = function_body,
            context       = function_context

        )

        if function_decl is not None:
            function_decl_codes.append(function_decl)


    # These are for functions used from other modules. Due to cyclic
    # dependencies, we cannot rely on those to be already created.
    for function_body in module.getCrossUsedFunctions():
        assert function_body.isCrossModuleUsed()

        function_decl = getFunctionDirectDecl(
            function_identifier = function_body.getCodeName(),
            closure_variables   = function_body.getClosureVariables(),
            file_scope          = getExportScopeCode(
                cross_module = function_body.isCrossModuleUsed()
            ),
            context             = Contexts.PythonFunctionDirectContext(
                parent   = context,
                function = function_body
            )
        )

        function_decl_codes.append(function_decl)

    for _identifier, code in sorted(iterItems(context.getHelperCodes())):
        function_body_codes.append(code)

    for _identifier, code in sorted(iterItems(context.getDeclarations())):
        function_decl_codes.append(code)

    function_body_codes = "\n\n".join(function_body_codes)
    function_decl_codes = "\n\n".join(function_decl_codes)

    template_values = getModuleValues(
        module_name         = module_name,
        module_identifier   = module.getCodeName(),
        codes               = codes.codes,
        function_decl_codes = function_decl_codes,
        function_body_codes = function_body_codes,
        temp_variables      = module.getTempVariables(),
        is_main_module      = module.isMainModule(),
        is_internal_module  = module.isInternalModule(),
        context             = context
    )

    if python_version >= 330:
        context.getConstantCode("__loader__")

    return template_values, context


def generateModuleCode(module_context, template_values):
    return getModuleCode(
        module_context  = module_context,
        template_values = template_values
    )


def generateHelpersCode(other_modules):
    calls_decl_code = getCallsDecls()

    loader_code = getMetapathLoaderBodyCode(other_modules)

    calls_body_code = getCallsCode()

    return calls_decl_code, calls_body_code + loader_code


def makeGlobalContext():
    return Contexts.PythonGlobalContext()


setExpressionDispatchDict(
    {
        "EXPRESSION_ATTRIBUTE_LOOKUP"               : generateAttributeLookupCode,
        "EXPRESSION_ATTRIBUTE_LOOKUP_SPECIAL"       : generateAttributeLookupSpecialCode,
        "EXPRESSION_BUILTIN_SLICE"                  : generateBuiltinSliceCode,
        "EXPRESSION_BUILTIN_HASH"                   : generateBuiltinHashCode,
        "EXPRESSION_BUILTIN_ID"                     : generateBuiltinIdCode,
        "EXPRESSION_BUILTIN_COMPILE"                : generateBuiltinCompileCode,
        "EXPRESSION_BUILTIN_EXECFILE"               : generateExecfileCode,
        "EXPRESSION_BUILTIN_EVAL"                   : generateEvalCode,
        "EXPRESSION_BUILTIN_EXEC"                   : generateEvalCode,
        "EXPRESSION_BUILTIN_ITER1"                  : generateBuiltinIter1Code,
        "EXPRESSION_BUILTIN_ITER2"                  : generateBuiltinIter2Code,
        "EXPRESSION_BUILTIN_NEXT1"                  : generateBuiltinNext1Code,
        "EXPRESSION_BUILTIN_NEXT2"                  : generateBuiltinNext2Code,
        "EXPRESSION_BUILTIN_SUM1"                   : generateBuiltinSum1Code,
        "EXPRESSION_BUILTIN_SUM2"                   : generateBuiltinSum2Code,
        "EXPRESSION_BUILTIN_TYPE1"                  : generateBuiltinType1Code,
        "EXPRESSION_BUILTIN_TYPE3"                  : generateBuiltinType3Code,
        "EXPRESSION_BUILTIN_IMPORT"                 : generateBuiltinImportCode,
        "EXPRESSION_BUILTIN_BOOL"                   : generateBuiltinBoolCode,
        "EXPRESSION_BUILTIN_BYTEARRAY"              : generateBuiltinBytearrayCode,
        "EXPRESSION_BUILTIN_INT"                    : generateBuiltinIntCode,
        "EXPRESSION_BUILTIN_LONG"                   : generateBuiltinLongCode,
        "EXPRESSION_BUILTIN_FLOAT"                  : generateBuiltinFloatCode,
        "EXPRESSION_BUILTIN_COMPLEX"                : generateBuiltinComplexCode,
        "EXPRESSION_BUILTIN_LEN"                    : generateBuiltinLenCode,
        "EXPRESSION_BUILTIN_STR"                    : generateBuiltinStrCode,
        "EXPRESSION_BUILTIN_UNICODE"                : generateBuiltinUnicodeCode,
        "EXPRESSION_BUILTIN_CHR"                    : generateBuiltinChrCode,
        "EXPRESSION_BUILTIN_ORD"                    : generateBuiltinOrdCode,
        "EXPRESSION_BUILTIN_BIN"                    : generateBuiltinBinCode,
        "EXPRESSION_BUILTIN_OCT"                    : generateBuiltinOctCode,
        "EXPRESSION_BUILTIN_HEX"                    : generateBuiltinHexCode,
        "EXPRESSION_BUILTIN_TUPLE"                  : generateBuiltinTupleCode,
        "EXPRESSION_BUILTIN_LIST"                   : generateBuiltinListCode,
        "EXPRESSION_BUILTIN_SET"                    : generateBuiltinSetCode,
        "EXPRESSION_BUILTIN_DICT"                   : generateBuiltinDictCode,
        "EXPRESSION_BUILTIN_LOCALS"                 : generateBuiltinLocalsCode,
        "EXPRESSION_BUILTIN_GLOBALS"                : generateBuiltinGlobalsCode,
        "EXPRESSION_BUILTIN_SUPER"                  : generateBuiltinSuperCode,
        "EXPRESSION_BUILTIN_ISINSTANCE"             : generateBuiltinIsinstanceCode,
        "EXPRESSION_BUILTIN_DIR1"                   : generateBuiltinDir1Code,
        "EXPRESSION_BUILTIN_VARS"                   : generateBuiltinVarsCode,
        "EXPRESSION_BUILTIN_HASATTR"                : generateBuiltinHasattrCode,
        "EXPRESSION_BUILTIN_GETATTR"                : generateBuiltinGetattrCode,
        "EXPRESSION_BUILTIN_SETATTR"                : generateBuiltinSetattrCode,
        "EXPRESSION_BUILTIN_OPEN"                   : generateBuiltinOpenCode,
        "EXPRESSION_BUILTIN_RANGE1"                 : generateBuiltinRange1Code,
        "EXPRESSION_BUILTIN_RANGE2"                 : generateBuiltinRange2Code,
        "EXPRESSION_BUILTIN_RANGE3"                 : generateBuiltinRange3Code,
        "EXPRESSION_BUILTIN_XRANGE1"                : generateBuiltinXrange1Code,
        "EXPRESSION_BUILTIN_XRANGE2"                : generateBuiltinXrange2Code,
        "EXPRESSION_BUILTIN_XRANGE3"                : generateBuiltinXrange3Code,
        "EXPRESSION_BUILTIN_MAKE_EXCEPTION"         : generateBuiltinMakeExceptionCode,
        "EXPRESSION_BUILTIN_REF"                    : generateBuiltinRefCode,
        "EXPRESSION_BUILTIN_EXCEPTION_REF"          : generateExceptionRefCode,
        "EXPRESSION_BUILTIN_ANONYMOUS_REF"          : generateBuiltinAnonymousRefCode,
        "EXPRESSION_CAUGHT_EXCEPTION_TYPE_REF"      : generateExceptionCaughtTypeCode,
        "EXPRESSION_CAUGHT_EXCEPTION_VALUE_REF"     : generateExceptionCaughtValueCode,
        "EXPRESSION_CAUGHT_EXCEPTION_TRACEBACK_REF" : generateExceptionCaughtTracebackCode,
        "EXPRESSION_CALL_EMPTY"                     : generateCallCode,
        "EXPRESSION_CALL_KEYWORDS_ONLY"             : generateCallCode,
        "EXPRESSION_CALL_NO_KEYWORDS"               : generateCallCode,
        "EXPRESSION_CALL"                           : generateCallCode,
        "EXPRESSION_CONSTANT_NONE_REF"              : generateConstantNoneReferenceCode,
        "EXPRESSION_CONSTANT_TRUE_REF"              : generateConstantTrueReferenceCode,
        "EXPRESSION_CONSTANT_FALSE_REF"             : generateConstantFalseReferenceCode,
        "EXPRESSION_CONSTANT_STR_REF"               : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_UNICODE_REF"           : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_BYTES_REF"             : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_INT_REF"               : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_LONG_REF"              : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_FLOAT_REF"             : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_COMPLEX_REF"           : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_ELLIPSIS_REF"          : generateConstantEllipsisReferenceCode,
        "EXPRESSION_CONSTANT_DICT_REF"              : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_DICT_EMPTY_REF"        : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_TUPLE_REF"             : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_LIST_REF"              : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_SET_REF"               : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_SLICE_REF"             : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_XRANGE_REF"            : generateConstantReferenceCode,
        "EXPRESSION_CONSTANT_TYPE_REF"              : generateConstantReferenceCode,
        "EXPRESSION_CONDITIONAL"                    : generateConditionalCode,
        "EXPRESSION_CONDITIONAL_OR"                 : generateConditionalAndOrCode,
        "EXPRESSION_CONDITIONAL_AND"                : generateConditionalAndOrCode,
        "EXPRESSION_COMPARISON"                     : generateComparisonExpressionCode,
        "EXPRESSION_COMPARISON_IS"                  : generateComparisonExpressionCode,
        "EXPRESSION_COMPARISON_IS_NOT"              : generateComparisonExpressionCode,
        "EXPRESSION_COMPARISON_IN"                  : generateComparisonExpressionCode,
        "EXPRESSION_COMPARISON_NOT_IN"              : generateComparisonExpressionCode,
        "EXPRESSION_COMPARISON_EXCEPTION_MATCH"     : generateComparisonExpressionCode,
        "EXPRESSION_DICT_OPERATION_GET"             : generateDictOperationGetCode,
        "EXPRESSION_DICT_OPERATION_IN"              : generateDictOperationInCode,
        "EXPRESSION_DICT_OPERATION_NOT_IN"          : generateDictOperationInCode,
        "EXPRESSION_FUNCTION_CREATION"              : generateFunctionCreationCode,
        "EXPRESSION_FUNCTION_CALL"                  : generateFunctionCallCode,
        "EXPRESSION_IMPORT_MODULE"                  : generateImportModuleCode,
        "EXPRESSION_IMPORT_MODULE_HARD"             : generateImportModuleHardCode,
        "EXPRESSION_IMPORT_NAME"                    : generateImportNameCode,
        "EXPRESSION_LIST_OPERATION_EXTEND"          : generateListOperationExtendCode,
        "EXPRESSION_LIST_OPERATION_POP"             : generateListOperationPopCode,
        "EXPRESSION_MODULE_FILE_ATTRIBUTE_REF"      : generateModuleFileAttributeCode,
        "EXPRESSION_MAKE_GENERATOR_OBJECT"          : generateMakeGeneratorObjectCode,
        "EXPRESSION_MAKE_COROUTINE_OBJECT"          : generateMakeCoroutineObjectCode,
        "EXPRESSION_MAKE_SET"                       : generateSetCreationCode,
        "EXPRESSION_MAKE_SET_LITERAL"               : generateSetLiteralCreationCode,
        "EXPRESSION_MAKE_TUPLE"                     : generateTupleCreationCode,
        "EXPRESSION_MAKE_LIST"                      : generateListCreationCode,
        "EXPRESSION_MAKE_DICT"                      : generateDictionaryCreationCode,
        "EXPRESSION_OPERATION_BINARY"               : generateOperationBinaryCode,
        "EXPRESSION_OPERATION_BINARY_ADD"           : generateOperationBinaryCode,
        "EXPRESSION_OPERATION_BINARY_MULT"          : generateOperationBinaryCode,
        "EXPRESSION_OPERATION_BINARY_INPLACE"       : generateOperationBinaryCode,
        "EXPRESSION_OPERATION_UNARY"                : generateOperationUnaryCode,
        "EXPRESSION_OPERATION_NOT"                  : generateOperationUnaryCode,
        "EXPRESSION_OUTLINE_BODY"                   : generateFunctionOutlineCode,
        "EXPRESSION_RETURNED_VALUE_REF"             : generateReturnedValueRefCode,
        "EXPRESSION_SUBSCRIPT_LOOKUP"               : generateSubscriptLookupCode,
        "EXPRESSION_SLICE_LOOKUP"                   : generateSliceLookupCode,
        "EXPRESSION_SET_OPERATION_UPDATE"           : generateSetOperationUpdateCode,
        "EXPRESSION_SIDE_EFFECTS"                   : generateSideEffectsCode,
        "EXPRESSION_SPECIAL_UNPACK"                 : generateSpecialUnpackCode,
        "EXPRESSION_TEMP_VARIABLE_REF"              : generateVariableReferenceCode,
        "EXPRESSION_VARIABLE_REF"                   : generateVariableReferenceCode,
        "EXPRESSION_YIELD"                          : generateYieldCode,
        "EXPRESSION_YIELD_FROM"                     : generateYieldFromCode,
        "EXPRESSION_SELECT_METACLASS"               : generateSelectMetaclassCode,
        "EXPRESSION_ASYNC_WAIT"                     : generateAsyncWaitCode,
        "EXPRESSION_ASYNC_ITER"                     : generateAsyncIterCode,
        "EXPRESSION_ASYNC_NEXT"                     : generateAsyncNextCode,
        "EXPRESSION_STRING_CONCATENATION"           : generateStringContenationCode,
        "EXPRESSION_BUILTIN_FORMAT"                 : generateBuiltinFormatCode,
        "EXPRESSION_BUILTIN_ASCII"                  : generateBuiltinAsciiCode,
    }
)

setStatementDispatchDict(
    {
        "STATEMENT_ASSIGNMENT_VARIABLE"      : generateAssignmentVariableCode,
        "STATEMENT_ASSIGNMENT_ATTRIBUTE"     : generateAssignmentAttributeCode,
        "STATEMENT_ASSIGNMENT_SUBSCRIPT"     : generateAssignmentSubscriptCode,
        "STATEMENT_ASSIGNMENT_SLICE"         : generateAssignmentSliceCode,
        "STATEMENT_DEL_VARIABLE"             : generateDelVariableCode,
        "STATEMENT_DEL_ATTRIBUTE"            : generateDelAttributeCode,
        "STATEMENT_DEL_SUBSCRIPT"            : generateDelSubscriptCode,
        "STATEMENT_DEL_SLICE"                : generateDelSliceCode,
        "STATEMENT_DICT_OPERATION_REMOVE"    : generateDictOperationRemoveCode,
        "STATEMENT_DICT_OPERATION_UPDATE"    : generateDictOperationUpdateCode,
        "STATEMENT_RELEASE_VARIABLE"         : generateVariableReleaseCode,
        "STATEMENT_EXPRESSION_ONLY"          : generateExpressionOnlyCode,
        "STATEMENT_RETURN"                   : generateReturnCode,
        "STATEMENT_GENERATOR_RETURN"         : generateGeneratorReturnCode,
        "STATEMENT_CONDITIONAL"              : generateBranchCode,
        "STATEMENT_TRY"                      : generateTryCode,
        "STATEMENT_PRINT_VALUE"              : generatePrintValueCode,
        "STATEMENT_PRINT_NEWLINE"            : generatePrintNewlineCode,
        "STATEMENT_IMPORT_STAR"              : generateImportStarCode,
        "STATEMENT_LIST_OPERATION_APPEND"    : generateListOperationAppendCode,
        "STATEMENT_SET_OPERATION_ADD"        : generateSetOperationAddCode,
        "STATEMENT_DICT_OPERATION_SET"       : generateDictOperationSetCode,
        "STATEMENT_LOOP"                     : generateLoopCode,
        "STATEMENT_LOOP_BREAK"               : generateLoopBreakCode,
        "STATEMENT_LOOP_CONTINUE"            : generateLoopContinueCode,
        "STATEMENT_RAISE_EXCEPTION"          : generateRaiseCode,
        "STATEMENT_RAISE_EXCEPTION_IMPLICIT" : generateRaiseCode,
        "STATEMENT_SPECIAL_UNPACK_CHECK"     : generateUnpackCheckCode,
        "STATEMENT_EXEC"                     : generateExecCode,
        "STATEMENT_LOCALS_DICT_SYNC"         : generateLocalsDictSyncCode,
        "STATEMENT_SET_LOCALS"               : generateSetLocalsCode,
        "STATEMENT_PRESERVE_FRAME_EXCEPTION" : generateFramePreserveExceptionCode,
        "STATEMENT_RESTORE_FRAME_EXCEPTION"  : generateFrameRestoreExceptionCode,
        "STATEMENT_PUBLISH_EXCEPTION"        : generateExceptionPublishCode
    }
)



#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Internal module

This is a container for helper functions that are shared across modules. It
may not exist, and is treated specially in code generation. This avoids to
own these functions to a random module.
"""


from nuitka.nodes.ModuleNodes import PythonInternalModule
from nuitka.SourceCodeReferences import fromFilename

internal_source_ref = fromFilename("internal").atInternal()

def once_decorator(func):
    """ Cache result of a function call without arguments.

    Used for all internal function accesses to become a singleton.

    Note: This doesn't much specific anymore, but we are not having
    this often enough to warrent re-use or generalization.

    """

    func.cached_value = None

    def replacement():
        if func.cached_value is None:
            func.cached_value = func()

        return func.cached_value

    return replacement


@once_decorator
def getInternalModule():
    """ Get the singleton internal module.

    """

    return PythonInternalModule()



#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Reformulation of boolean and/or expressions.

Consult the developer manual for information. TODO: Add ability to sync
source code comments with developer manual sections.

"""

from nuitka.nodes.ConditionalNodes import (
    ExpressionConditionalAND,
    ExpressionConditionalOR
)
from nuitka.nodes.OperatorNodes import ExpressionOperationNOT

from .Helpers import buildNode, buildNodeList, getKind


def buildBoolOpNode(provider, node, source_ref):
    bool_op = getKind(node.op)

    if bool_op == "Or":
        # The "or" may be short circuit and is therefore not a plain operation.
        values = buildNodeList(provider, node.values, source_ref)

        for value in values[:-1]:
            value.setCompatibleSourceReference(values[-1].getSourceReference())

        source_ref = values[-1].getSourceReference()

        return buildOrNode(
            values     = values,
            source_ref = source_ref
        )

    elif bool_op == "And":
        # The "and" may be short circuit and is therefore not a plain operation.
        values = buildNodeList(provider, node.values, source_ref)

        for value in values[:-1]:
            value.setCompatibleSourceReference(values[-1].getSourceReference())

        source_ref = values[-1].getSourceReference()

        return buildAndNode(
            values     = values,
            source_ref = source_ref
        )
    elif bool_op == "Not":
        # The "not" is really only a unary operation and no special.
        return ExpressionOperationNOT(
            operand    = buildNode(provider, node.operand, source_ref),
            source_ref = source_ref
        )
    else:
        assert False, bool_op


def buildOrNode(values, source_ref):
    values = list(values)

    result = values.pop()

    # When we encounter, "or", we expect it to be at least two values.
    assert values

    while values:
        result = ExpressionConditionalOR(
            left       = values.pop(),
            right      = result,
            source_ref = source_ref
        )

    return result


def buildAndNode(values, source_ref):
    values = list(values)

    result = values.pop()

    # Unlike "or", for "and", this is used with only one value.

    while values:
        result = ExpressionConditionalAND(
            left       = values.pop(),
            right      = result,
            source_ref = source_ref
        )

    return result



#!/usr/bin/env python
#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

""" Sort import statements using isort for Nuitka source.

"""

from __future__ import print_function

import os
import subprocess
import sys

# Unchanged, running from checkout, use the parent directory, the nuitka
# package ought be there.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            ".."
        )
    )
)

from nuitka.tools.Basics import goHome, addPYTHONPATH, setupPATH # isort:skip
from nuitka.tools.ScanSources import scanTargets # isort:skip


def main():
    goHome()

    # So PyLint finds nuitka package.
    addPYTHONPATH(os.getcwd())

    target_files = []
    for filename in scanTargets(["nuitka", "bin"]):
        target_files.append(filename)

    target_files.append("nuitka/build/SingleExe.scons")

    setupPATH()
    subprocess.check_call(
        [
            "isort",
            "-ot",
            "-m3",
            "-ns",
            "__init__.py"
        ] + target_files
    )

if __name__ == "__main__":
    main()



#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" PyLint handling for Nuitka.

Our usage of PyLint also works around a few issues that PyLint
has.

"""

from __future__ import print_function

import os
import subprocess
import sys


def checkVersion():
    pylint_version = subprocess.check_output(
        ["pylint", "--version"],
        stderr = open(os.devnull, 'w')
    )

    pylint_version = pylint_version.split(b"\n")[0].split()[-1]

    if pylint_version < b"1.4.4":
        sys.exit("Error, needs PyLint 1.4.4 or higher.")


# Disabled globally:
#
# W0232: Class has no __init__ method
# Who cares, we are using overrides that don't need to change object init a lot
# and rarely ever made a mistake with forgetting to call used __init__ of the
# parent.
#
# I0011: Locally disabling W....
# Strange one anyway, we want to locally disable stuff. And that just makes it
# a different warning. Amazing. Luckily we can decide to ignore that globally
# then.
#
# I0012: Locally enabling W....
# Sure, we disabled it for a block, and re-enabled it then.
#
# C0326: No space allowed...
# Our spaces before keyword argument calls are not allowed, and this is
# not possible to distinguish.
#
# C0330: Wrong hanging line indentation
# No it's not wrong.
#
# C1001: Old-style class defined.
# Yes, so what, why care.
#
# E1120 / E1123: ....
# Constructor call checks frequently fail miserably, so this is full of
# mysterious false alarms, while it's unlikely to help much.
#
# E1103: Instance of 'x' has no 'y' member but some types could not be inferred
# Rarely is this any help, but it's full of false alarms.
#
# W0632: Possible unbalanced tuple unpacking with sequence defined at ...
# It's not really good at guessing these things.
#
# W1504: Using type() instead of isinstance() for a typecheck.
# Nuitka is all about exact type checks, so this doesn't apply
#
# C0123: Using type() instead of isinstance() for a typecheck.
# Nuitka is all about exact type checks, so this doesn't apply
#
# C0413: Import "..." should be placed at the top of the module
# There is no harm to this and imports are deal with by isort binary.
#
# C0411: external import "external" comes before "local"
# There is no harm to this and imports are deal with by isort binary.
#
# R0204: Redefinition of var type from x to y
# I do this all the time, e.g. to convert str to unicode, or list to string.

default_pylint_options = """\
--rcfile=/dev/null
--disable=I0011,I0012,W0232,C0326,C0330,C1001,E1103,W0632,W1504,C0123,C0413,C0411,R0204
--msg-template="{path}:{line} {msg_id} {obj} {msg}"
--reports=no
--persistent=no
--method-rgx=[a-z_][a-zA-Z0-9_]{2,40}$
--module-rgx=.*
--function-rgx=.*
--variable-rgx=.*
--argument-rgx=.*
--dummy-variables-rgx=_.*|trace_collection
--const-rgx=.*
--max-line-length=120
--no-docstring-rgx=.*
--max-module-lines=5000
--min-public-methods=0
--max-public-methods=100
--max-args=10
--max-parents=10
--max-nested-blocks=10
--max-bool-expr=10
--enable=useless-suppression""".split('\n')

our_exit_code = 0

def executePyLint(filename, show_todos, verbose):
    if verbose:
        print("Checking", filename, "...")

    pylint_options = list(default_pylint_options)
    if not show_todos:
        pylint_options.append("--notes=")

    # This is kind of a singleton module, pylint: disable=W0603
    global our_exit_code

    extra_options = os.environ.get("PYLINT_EXTRA_OPTIONS", "").split()
    if "" in extra_options:
        extra_options.remove("")
    command = ["pylint"] + pylint_options + extra_options + [filename]

    process = subprocess.Popen(
        args   = command,
        stdout = subprocess.PIPE,
        stderr = subprocess.STDOUT,
        shell  = False
    )

    stdout, stderr = process.communicate()
    _exit_code = process.returncode

    assert not stderr, stderr

    if stdout:
        stdout = stdout.replace("\r\n", '\n')

        # Remove hard to disable error line given under Windows.
        lines = stdout.split(b"\n")
        try:
            error_line = lines.index(
                b"No config file found, using default configuration"
            )
            del lines[error_line]
            del lines[error_line]
        except ValueError:
            pass

        lines = [
            line.decode()
            for line in
            lines
        ]

        lines = [
            line
            for line in
            lines
            if "Unable to import 'resource'" not in line
        ]

        # If we filtered everything away, remove the leading file name report.
        if len(lines) == 1:
            assert lines[0].startswith("*****")
            lines = []

        for line in lines:
            print(line)

        if lines:
            our_exit_code = 1

    sys.stdout.flush()
